function [A_id_cell,U_id_cell,est_err_cell,sol_found] = ...
    main_corenet_nopa(Meas_mtx,B_vct,sign_constr_list,...
    perc_edge,max_iter_num)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% F. Amato, C. Cosentino,  F. Montefusco
% Biomechatronics Laboratory
% Universita' Magna Graecia di Catanzaro
% Campus Universitario "Salvatore Venuta"
% viale Europa, loc. Germaneto
% 88100 Catanzaro ITALY
% Tel:+39-0961-369-4196
% fax: +39-0961-369-4090
% e-mail: f.montefusco@unicz.it
%
%
% Let us assume that a biological network is represented
% by a stable dynamical linear system, in the form dot(x)= Ax +Bu,
% The function allows to identify  the connectivity matrix A and exogenous
% perturbation vector B from the experimental measurements
% via convex optimization problem in the form of LMIs.
% The initialization network is identified by imposing only selfloops on 
% every vertex.
% Subsequently, new edges are added step-by-step, from the coefficients 
% with the highest weights, including those derived by a priori 
% information. The iterative procedure stops when the number of the 
% iterations is equal to that maximum.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUT VARIABLES
%
%
% Meas_mtx:     Measurements matrix, each column contains the measurements 
%               for a node
%
% B_vct:        Vector that models the effect of external
%               perturbations 
%
% sign_constr_list: Matrix with 3 columns.
%             Constraints to be imposed on the matrix coefficients
%             first column --> row-index of the constrained coefficient
%             second column --> column-index of the constrained coefficient
%             third column --> sign of the constrained coefficient
%             
% perc_edge:    Coefficient that determinates the number of edges that are
%               selected and added to the network at each step
%
% max_iter_num: maximum number of iterations
%
% OUTPUT VARIABLES
%
% A_id_cell:    Cell array that contains the identified connectivity  
%               matrices at each step 
%
% U_id_cell:    Cell array that contains the identified input vectors at  
%               each step
%
% est_err_cell: Cell array that contains the norm of vector of residuals    
%               generated by the identified model at each step 
%
% sol_found     Flag true, if a solution has been found
%                
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin==4
    % Default value of max_iter_num
    max_iter_num=10;
 
elseif nargin~=5
    % Wrong number of input arguments
    error('Wrong number of input arguments');
end

% determine the number of time points and nodes, respectively
[n_time_pts,n_nodes]=size(Meas_mtx);

% iteration counter
iter_num=1;

% number of the identifications
ident_num=1;

% number of the added edges
n_link_added=0;

% define matrix which contains the added edges
if isempty(sign_constr_list)    
    Constr_mtx=[];
else
    Constr_mtx=sign_constr_list(:,1:2);
end

% define a list of coefficents that determine a not 
% solution of the lmi problem (Unstb_list) 
Unstbl_list=[];

%build the list of coefficients to be nullified
null_constr_list=[];
if not(isempty(sign_constr_list))
    for idx1=1:n_nodes
        for idx2=[1:idx1-1,idx1+1:n_nodes]
            % Check if the element (idx1,idx2) is contained in the
            % sign constr list
            if not(sum(double(idx1==sign_constr_list(:,1)& idx2==sign_constr_list(:,2))))
                null_constr_list_old=null_constr_list;
                null_constr_list=[null_constr_list_old; idx1 idx2];
                clear null_constr_list_old
            end
        end
    end
else
    for idx1=1:n_nodes
        for idx2=[1:idx1-1,idx1+1:n_nodes]
            null_constr_list_old=null_constr_list;
            null_constr_list=[null_constr_list_old; idx1 idx2];
            clear null_constr_list_old
        end
    end
end

%Find the b-coefficients to be nullified and to be constrained
null_b_list=[];
if not(isempty(B_vct))
   for idx1=1:n_nodes
       if not(B_vct(idx1))
           null_b_list_old=null_b_list;
           null_b_list=[null_b_list_old; idx1 ];
           clear null_b_list_old
       else
           sign_constr_list_old=sign_constr_list;
           sign_constr_list=[sign_constr_list_old;idx1 n_nodes+1 sign(B_vct(idx1))];
           clear sign_constr_list_old
       end
   end
end
          
%first identification:isolated network
[A_id,U_id,norm_est_err_dy] = ...
    lmi_sign_constr_ex(Meas_mtx,sign_constr_list,null_constr_list,null_b_list);

%store the identification
A_id_cell{ident_num}=A_id;
U_id_cell{ident_num}=U_id;
est_err_cell{ident_num}=norm_est_err_dy;


% identification without deletion of the coefficients
[A_id_temp,U_id_temp,norm_est_err_dy_temp,sol_found_temp] = ...
    lmi_sign_constr_full_ex(Meas_mtx,sign_constr_list,null_b_list);
sol_found=sol_found_temp;
if not(sol_found_temp)
    return
end

% Compute the normalized A matrix
Normal_A_id=zeros(n_nodes,n_nodes);
for idx1=1:n_nodes
    row_norm=norm(A_id_temp(idx1,:));
    for idx2=1:n_nodes
        % Normalization of each coefficient with respect to the values on
        % the same row and column
        Normal_A_id(idx1,idx2)=abs(A_id_temp(idx1,idx2))/sqrt(row_norm*norm(A_id_temp(:,idx2)));
    end
end

% Don't take account the diagonal coefficients and the prior constraints to 
% generate the edges list to be add added  
for idx2=1:n_nodes
    Normal_A_id(idx2,idx2)=0;
end
if not(isempty(Constr_mtx))
    for idx1=1:size(Constr_mtx,1)
        Normal_A_id(Constr_mtx(idx1,1),Constr_mtx(idx1,2))=0;
    end
end
% Compute the edge ranking matrix (Constr_list)
[idx_row idx_col val_coef]=find(Normal_A_id);
[val_coef_sort idx_sort]=sort(val_coef,'descend');
Constr_list=[idx_row(idx_sort) idx_col(idx_sort) val_coef_sort];

num_it_ins_all_nodes=0;% flag, if >0, then all the edges were added
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%     Algorithm for adding  the coefficients  %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

while   iter_num<=max_iter_num  
    
    % compute the mu_k function, a decreasing function of k (k=iter_num)
    n_link_added_id=ceil(ceil(n_nodes/perc_edge)*n_nodes/(iter_num+1));
    
    % number of edges to be added at k iteration
    n_link_added=n_link_added_id+n_link_added;
    
    % define the new list of the edges to be added (new_Constr_mtx) 
    if n_link_added<size(Constr_list,1)
        new_Constr_mtx=Constr_list(1:n_link_added,1:2);
    else
        num_it_ins_all_nodes=1+num_it_ins_all_nodes;
        new_Constr_mtx=Constr_list;
    end
    if num_it_ins_all_nodes>1
        break
    end
    
    % Check if the elements of the new_Constr_mtx has been already added
    % or determine a not solution of the lmi problem (added to Unstb_list)
    temp_Constr_mtx=[];
    for idx1=1:size(new_Constr_mtx,1)
        in_constr_list=false;
        for idx2=1:size(Constr_mtx,1)
            % Compare with each row of the constr_list
            if  prod(double(Constr_mtx(idx2,:)==[new_Constr_mtx(idx1,1) new_Constr_mtx(idx1,2) ]))
                in_constr_list=true;
            end
        end
        in_unstbl_list=false;
        for idx2=1:size(Unstbl_list,1)
            % Compare with each row of the constr_list
            if  prod(double(Unstbl_list(idx2,:)==[new_Constr_mtx(idx1,1) new_Constr_mtx(idx1,2) ]))
                in_unstbl_list=true;
            end
        end
        if not(in_constr_list) && not(in_unstbl_list)
            temp_Constr_mtx_old=temp_Constr_mtx;
            temp_Constr_mtx=[temp_Constr_mtx_old;  new_Constr_mtx(idx1,1) new_Constr_mtx(idx1,2) ];
            clear temp_Constr_mtx_old
        end  
    end
    if not(isempty(temp_Constr_mtx))
        %delete from the null list the elements (edges) to be added
        temp_null_constr_list=null_constr_list;
        for idx1=1:size(temp_Constr_mtx)
            % vector containing all zeros except one (the element/edge to be
            % inserted)
            pos_null_list=(temp_Constr_mtx(idx1,1)==temp_null_constr_list(:,1)& ...
                temp_Constr_mtx(idx1,2)==temp_null_constr_list(:,2));
            temp_null_constr_list(pos_null_list(true(size(pos_null_list,1),1)),:)=[];
        end
        % identification by using the new list of
        % coefficient constraints
        [A_id,U_id,norm_est_err_dy,sol_found] = ...
            lmi_sign_constr_ex(Meas_mtx,sign_constr_list,temp_null_constr_list,null_b_list);
        if not(sol_found)
            % determine the elements that determine a not solution of the
            % lmi problem
            Unstbl_list_old=Unstbl_list;
            Unstbl_list=[Unstbl_list_old; temp_Constr_mtx];
            clear Unstbl_list_old
            unstbl_link_added=0; % flag, true if for at least one edge of the 
                                 % Unstbl_list a solution has been found
           
            % define new null list (delete an edge at a time)
            % and then check if a solution has been found
            temp_unstbl_null_constr_list=null_constr_list;
            for idx=1:size(Unstbl_list,1)
                % vector containing all zeros except one (the element/edge to be
                % added)
                pos_null_list=(Unstbl_list(idx1,1)==temp_unstbl_null_constr_list(:,1)& ...
                    Unstbl_list(idx1,2)==temp_unstbl_null_constr_list(:,2));
                
                % vector containing the indices of edge to be deleted from the
                % null list
                check_edge=temp_unstbl_null_constr_list(pos_null_list(true(size(pos_null_list,1),1)),:);
                
                % define the new null list
                temp_unstbl_null_constr_list(pos_null_list(true(size(pos_null_list,1),1)),:)=[];
                
                % identification by using the new list of
                % coefficient constraints
                [A_id_1,U_id_1,norm_est_err_dy_1,sol_found_1] = ...
                    lmi_sign_constr_ex(Meas_mtx,sign_constr_list,temp_unstbl_null_constr_list,null_b_list);
                if sol_found_1
                    % for the added edge a solution of the problem has been
                    % found
                    unstbl_link_added=1;
                    % Add it (the edge) in the matrix  of the added edges
                    % (update Constr_mtx)
                    Constr_mtx_old=Constr_mtx;
                    Constr_mtx=[Constr_mtx_old;Unstbl_list(idx,:)];
                    clear Constr_mtx_old
                else
                    % update the new null list
                    temp_unstbl_null_constr_list_old=temp_unstbl_null_constr_list;
                    temp_unstbl_null_constr_list=[temp_unstbl_null_constr_list_old;check_edge];
                end
            end
            if unstbl_link_added
                % update the Unstable list, for at least one edge of the 
                % Unstbl_list a solution has been found
                Unstbl_list=[];
                
                % update the null list
                null_constr_list=temp_unstbl_null_constr_list; 
                
                % identification by using the edges that determine a
                % solution of the lmi problem
                [A_id,U_id,norm_est_err_dy,sol_found] = ...
                    lmi_sign_constr_ex(Meas_mtx,sign_constr_list,null_constr_list,null_b_list);
                ident_num=1+ident_num;% increase the number of the identifications
                % store the identificated variables
                A_id_cell{ident_num}=A_id;
                U_id_cell{ident_num}=U_id;
                est_err_cell{ident_num}=norm_est_err_dy;
                
            end
            clear temp_unstbl_null_constr_list
        else
            % the solution has been found, then redefine the Unstb_list
            Unstbl_list=[];
            ident_num=1+ident_num;% increase the number of the identifications
            % update the null list
            null_constr_list=temp_null_constr_list;
            
            % store the identified variables
            A_id_cell{ident_num}=A_id;
            U_id_cell{ident_num}=U_id;
            est_err_cell{ident_num}=norm_est_err_dy;
            
            
            % update the list of added edges
            Constr_mtx_old=Constr_mtx;
            Constr_mtx=[Constr_mtx_old;temp_Constr_mtx];
        end
        clear temp_Constr_mtx new_Constr_mtx  temp_null_constr_list
    end
    iter_num=iter_num+1;
end

